@@ -144,8 +144,6 @@ static inline void conn_ctrl_init(struct connection *conn)
 static inline void conn_ctrl_close(struct connection *conn)
 {
 	if ((conn->flags & (CO_FL_XPRT_READY|CO_FL_CTRL_READY)) == CO_FL_CTRL_READY) {
-		if ((conn->flags & (CO_FL_WANT_DRAIN | CO_FL_SOCK_RD_SH)) == CO_FL_WANT_DRAIN)
-			conn_sock_drain(conn);
 		fd_delete(conn->handle.fd);
 		conn->handle.fd = DEAD_FD_MAGIC;
 		conn->flags &= ~CO_FL_CTRL_READY;
@@ -244,8 +242,6 @@ static inline void conn_xprt_shutw_hard(struct connection *c)
 /* shut read */
 static inline void cs_shutr(struct conn_stream *cs, enum cs_shr_mode mode)
 {
-	if (cs->flags & CS_FL_SHR)
-		return;
 
 	/* clean data-layer shutdown */
 	if (cs->conn->mux && cs->conn->mux->shutr)
@@ -256,8 +252,6 @@ static inline void cs_shutr(struct conn_stream *cs, enum cs_shr_mode mode)
 /* shut write */
 static inline void cs_shutw(struct conn_stream *cs, enum cs_shw_mode mode)
 {
-	if (cs->flags & CS_FL_SHW)
-		return;
 
 	/* clean data-layer shutdown */
 	if (cs->conn->mux && cs->conn->mux->shutw)
@@ -270,6 +264,7 @@ static inline void cs_close(struct conn_stream *cs)
 {
 	cs_shutw(cs, CS_SHW_SILENT);
 	cs_shutr(cs, CS_SHR_RESET);
+	cs->flags = CS_FL_NONE;
 }
 
 /* sets CS_FL_ERROR or CS_FL_ERR_PENDING on the cs */
@@ -317,7 +312,7 @@ static inline void cs_init(struct conn_stream *cs, struct connection *conn)
  * is about to be reused. It also leaves the addresses untouched, which makes
  * it usable across connection retries to reset a connection to a known state.
  */
-static inline void conn_init(struct connection *conn)
+static inline void conn_init(struct connection *conn, void *target)
 {
 	conn->obj_type = OBJ_TYPE_CONN;
 	conn->flags = CO_FL_NONE;
@@ -327,13 +322,12 @@ static inline void conn_init(struct connection *conn)
 	conn->send_proxy_ofs = 0;
 	conn->handle.fd = DEAD_FD_MAGIC;
 	conn->err_code = CO_ER_NONE;
-	conn->target = NULL;
+	conn->target = target;
 	conn->destroy_cb = NULL;
 	conn->proxy_netns = NULL;
 	MT_LIST_INIT(&conn->list);
 	LIST_INIT(&conn->session_list);
 	conn->subs = NULL;
-	conn->idle_time = 0;
 	conn->src = NULL;
 	conn->dst = NULL;
 	conn->proxy_authority = NULL;
@@ -347,13 +341,27 @@ static inline void conn_set_owner(struct connection *conn, void *owner, void (*c
 	conn->destroy_cb = cb;
 }
 
+
+/* Mark the connection <conn> as private and remove it from the available connection list */
+static inline void conn_set_private(struct connection *conn)
+{
+	if (!(conn->flags & CO_FL_PRIVATE)) {
+		conn->flags |= CO_FL_PRIVATE;
+
+		if (obj_type(conn->target) == OBJ_TYPE_SERVER)
+			srv_del_conn_from_list(__objt_server(conn->target), conn);
+	}
+}
+
 /* Allocates a struct sockaddr from the pool if needed, assigns it to *sap and
  * returns it. If <sap> is NULL, the address is always allocated and returned.
  * if <sap> is non-null, an address will only be allocated if it points to a
  * non-null pointer. In this case the allocated address will be assigned there.
- * In both situations the new pointer is returned.
+ * If <orig> is non-null and <len> positive, the address in <sa> will be copied
+ * into the allocated address. In both situations the new pointer is returned.
  */
-static inline struct sockaddr_storage *sockaddr_alloc(struct sockaddr_storage **sap)
+static inline struct sockaddr_storage *
+sockaddr_alloc(struct sockaddr_storage **sap, const struct sockaddr_storage *orig, socklen_t len)
 {
 	struct sockaddr_storage *sa;
 
@@ -361,6 +369,8 @@ static inline struct sockaddr_storage *sockaddr_alloc(struct sockaddr_storage **
 		return *sap;
 
 	sa = pool_alloc(pool_head_sockaddr);
+	if (sa && orig && len > 0)
+		memcpy(sa, orig, len);
 	if (sap)
 		*sap = sa;
 	return sa;
@@ -382,13 +392,16 @@ static inline void sockaddr_free(struct sockaddr_storage **sap)
  * connection is returned on success, NULL on failure. The connection must
  * be released using pool_free() or conn_free().
  */
-static inline struct connection *conn_new()
+static inline struct connection *conn_new(void *target)
 {
 	struct connection *conn;
 
 	conn = pool_alloc(pool_head_connection);
-	if (likely(conn != NULL))
-		conn_init(conn);
+	if (likely(conn != NULL)) {
+		conn_init(conn, target);
+		if (obj_type(target) == OBJ_TYPE_SERVER)
+			srv_use_conn(__objt_server(target), conn);
+	}
 	return conn;
 }
 
@@ -408,7 +421,7 @@ static inline void cs_free(struct conn_stream *cs)
  * to the mux's stream list on success, then returned. On failure, nothing is
  * allocated and NULL is returned.
  */
-static inline struct conn_stream *cs_new(struct connection *conn)
+static inline struct conn_stream *cs_new(struct connection *conn, void *target)
 {
 	struct conn_stream *cs;
 
@@ -417,12 +430,11 @@ static inline struct conn_stream *cs_new(struct connection *conn)
 		return NULL;
 
 	if (!conn) {
-		conn = conn_new();
+		conn = conn_new(target);
 		if (unlikely(!conn)) {
 			cs_free(cs);
 			return NULL;
 		}
-		conn_init(conn);
 	}
 
 	cs_init(cs, conn);
@@ -456,12 +468,14 @@ static inline void conn_force_unsubscribe(struct connection *conn)
 /* Releases a connection previously allocated by conn_new() */
 static inline void conn_free(struct connection *conn)
 {
-	/* Remove ourself from the session's connections list, if any. */
-	if (!LIST_ISEMPTY(&conn->session_list)) {
-		struct session *sess = conn->owner;
-		if (conn->flags & CO_FL_SESS_IDLE)
-			sess->idle_conns--;
-		session_unown_conn(sess, conn);
+	/* If the connection is owned by the session, remove it from its list
+	 */
+	if (LIST_ADDED(&conn->session_list)) {
+		session_unown_conn(conn->owner, conn);
+	}
+	else if (!(conn->flags & CO_FL_PRIVATE)) {
+		if (obj_type(conn->target) == OBJ_TYPE_SERVER)
+			srv_del_conn_from_list(__objt_server(conn->target), conn);
 	}
 
 	sockaddr_free(&conn->src);
@@ -483,23 +497,7 @@ static inline void conn_free(struct connection *conn)
 	if (conn->ctx != NULL && conn->mux == NULL)
 		*(void **)conn->ctx = NULL;
 
-	/* The connection is currently in the server's idle list, so tell it
-	 * there's one less connection available in that list.
-	 */
-	if (conn->idle_time > 0) {
-		struct server *srv = __objt_server(conn->target);
-		_HA_ATOMIC_SUB(&srv->curr_idle_conns, 1);
-		_HA_ATOMIC_SUB(conn->flags & CO_FL_SAFE_LIST ? &srv->curr_safe_nb : &srv->curr_idle_nb, 1);
-		_HA_ATOMIC_SUB(&srv->curr_idle_thr[tid], 1);
-	} else {
-		struct server *srv = objt_server(conn->target);
-
-		if (srv)
-			_HA_ATOMIC_SUB(&srv->curr_used_conns, 1);
-	}
-
 	conn_force_unsubscribe(conn);
-	MT_LIST_DEL((struct mt_list *)&conn->list);
 	pool_free(pool_head_connection, conn);
 }
 
@@ -538,13 +536,13 @@ static inline int conn_get_src(struct connection *conn)
 	if (conn->flags & CO_FL_ADDR_FROM_SET)
 		return 1;
 
-	if (!conn_ctrl_ready(conn) || !conn->ctrl->get_src)
+	if (!conn_ctrl_ready(conn) || !conn->ctrl->fam->get_src)
 		return 0;
 
-	if (!sockaddr_alloc(&conn->src))
+	if (!sockaddr_alloc(&conn->src, NULL, 0))
 		return 0;
 
-	if (conn->ctrl->get_src(conn->handle.fd, (struct sockaddr *)conn->src,
+	if (conn->ctrl->fam->get_src(conn->handle.fd, (struct sockaddr *)conn->src,
 	                        sizeof(*conn->src),
 	                        obj_type(conn->target) != OBJ_TYPE_LISTENER) == -1)
 		return 0;
@@ -561,13 +559,13 @@ static inline int conn_get_dst(struct connection *conn)
 	if (conn->flags & CO_FL_ADDR_TO_SET)
 		return 1;
 
-	if (!conn_ctrl_ready(conn) || !conn->ctrl->get_dst)
+	if (!conn_ctrl_ready(conn) || !conn->ctrl->fam->get_dst)
 		return 0;
 
-	if (!sockaddr_alloc(&conn->dst))
+	if (!sockaddr_alloc(&conn->dst, NULL, 0))
 		return 0;
 
-	if (conn->ctrl->get_dst(conn->handle.fd, (struct sockaddr *)conn->dst,
+	if (conn->ctrl->fam->get_dst(conn->handle.fd, (struct sockaddr *)conn->dst,
 	                        sizeof(*conn->dst),
 	                        obj_type(conn->target) != OBJ_TYPE_LISTENER) == -1)
 		return 0;
@@ -688,7 +686,6 @@ static inline const char *conn_err_code_str(struct connection *c)
 
 	case CO_ER_CIP_EMPTY:     return "Connection closed while waiting for NetScaler Client IP header";
 	case CO_ER_CIP_ABORT:     return "Connection error while waiting for NetScaler Client IP header";
-	case CO_ER_CIP_TIMEOUT:   return "Timeout while waiting for a NetScaler Client IP header";
 	case CO_ER_CIP_TRUNCATED: return "Truncated NetScaler Client IP header received";
 	case CO_ER_CIP_BAD_MAGIC: return "Received an invalid NetScaler Client IP magic number";
 	case CO_ER_CIP_BAD_PROTO: return "Received an unhandled protocol in the NetScaler Client IP header";
@@ -707,7 +704,6 @@ static inline const char *conn_err_code_str(struct connection *c)
 	case CO_ER_SSL_HANDSHAKE_HB: return "SSL handshake failure after heartbeat";
 	case CO_ER_SSL_KILLED_HB: return "Stopped a TLSv1 heartbeat attack (CVE-2014-0160)";
 	case CO_ER_SSL_NO_TARGET: return "Attempt to use SSL on an unknown target (internal error)";
-	case CO_ER_SSL_EARLY_FAILED: return "Server refused early data";
 
 	case CO_ER_SOCKS4_SEND:    return "SOCKS4 Proxy write error during handshake";
 	case CO_ER_SOCKS4_RECV:    return "SOCKS4 Proxy read error during handshake";
