@@ -37,7 +37,12 @@ extern const struct name_desc stat_fields[];
 extern const struct name_desc info_fields[];
 extern const char *stat_status_codes[];
 extern struct applet http_stats_applet;
-extern THREAD_LOCAL struct field info[];
+
+
+struct htx;
+int stats_putchk(struct channel *chn, struct htx *htx, struct buffer *chk);
+
+int stats_dump_one_line(const struct field *stats, size_t stats_count, struct appctx *appctx);
 
 int stats_fill_info(struct field *info, int len);
 int stats_fill_fe_stats(struct proxy *px, struct field *stats, int len);
@@ -115,6 +120,13 @@ static inline struct field mkf_flt(uint32_t type, double value)
 	return f;
 }
 
+#define MK_STATS_PROXY_DOMAIN(px_cap) \
+	((px_cap) << STATS_PX_CAP | STATS_DOMAIN_PROXY)
+
+int stats_allocate_proxy_counters(struct proxy *px);
+
+void stats_register_module(struct stats_module *m);
+
 #endif /* _HAPROXY_STATS_H */
 
 /*
