@@ -22,28 +22,6 @@
 #include <haproxy/tools.h>
 
 
-/* Check an action ruleset validity. It returns the number of error encountered
- * andd err_code is updated if a warning is emitted.
- */
-int check_action_rules(struct list *rules, struct proxy *px, int *err_code)
-{
-	struct act_rule *rule;
-	char *errmsg = NULL;
-	int err = 0;
-
-	list_for_each_entry(rule, rules, list) {
-		if (rule->check_ptr && !rule->check_ptr(rule, px, &errmsg)) {
-			ha_alert("Proxy '%s': %s.\n", px->id, errmsg);
-			err++;
-		}
-
-		free(errmsg);
-		errmsg = NULL;
-	}
-
-	return err;
-}
-
 /* Find and check the target table used by an action track-sc*. This
  * function should be called during the configuration validity check.
  *
@@ -90,13 +68,30 @@ int check_trk_action(struct act_rule *rule, struct proxy *px, char **err)
 		 */
 	}
 
-	if (rule->from == ACT_F_TCP_REQ_CNT && (px->cap & PR_CAP_FE) && !px->tcp_req.inspect_delay &&
-	    !(rule->arg.trk_ctr.expr->fetch->val & SMP_VAL_FE_SES_ACC)) {
-		ha_warning("config : %s '%s' : a 'tcp-request content track-sc*' rule explicitly depending on request"
-			   " contents without any 'tcp-request inspect-delay' setting."
-			   " This means that this rule will randomly find its contents. This can be fixed by"
-			   " setting the tcp-request inspect-delay.\n",
-			   proxy_type_str(px), px->id);
+	if (rule->from == ACT_F_TCP_REQ_CNT && (px->cap & PR_CAP_FE)) {
+		if (!px->tcp_req.inspect_delay && !(rule->arg.trk_ctr.expr->fetch->val & SMP_VAL_FE_SES_ACC)) {
+			ha_warning("config : %s '%s' : a 'tcp-request content track-sc*' rule explicitly depending on request"
+				   " contents without any 'tcp-request inspect-delay' setting."
+				   " This means that this rule will randomly find its contents. This can be fixed by"
+				   " setting the tcp-request inspect-delay.\n",
+				   proxy_type_str(px), px->id);
+		}
+
+		/* The following warning is emitted because HTTP multiplexers are able to catch errors
+		 * or timeouts at the session level, before instantiating any stream.
+		 * Thus the tcp-request content ruleset will not be evaluated in such case. It means,
+		 * http_req and http_err counters will not be incremented as expected, even if the tracked
+		 * counter does not use the request content. To track invalid requests it should be
+		 * performed at the session level using a tcp-request session rule.
+		 */
+		if (px->mode == PR_MODE_HTTP &&
+		    !(rule->arg.trk_ctr.expr->fetch->use & (SMP_USE_L6REQ|SMP_USE_HRQHV|SMP_USE_HRQHP|SMP_USE_HRQBO)) &&
+		    (!rule->cond || !(rule->cond->use & (SMP_USE_L6REQ|SMP_USE_HRQHV|SMP_USE_HRQHP|SMP_USE_HRQBO)))) {
+			ha_warning("config : %s '%s' : a 'tcp-request content track-sc*' rule not depending on request"
+				   " contents for an HTTP frontend should be executed at the session level, using a"
+				   " 'tcp-request session' rule (mandatory to track invalid HTTP requests).\n",
+				   proxy_type_str(px), px->id);
+		}
 	}
 
 	return 1;
@@ -138,12 +133,6 @@ int act_resolution_cb(struct dns_requester *requester, struct dns_nameserver *na
 	return 0;
 }
 
-/*
- * Do resolve error management callback
- * returns:
- *  0 if we can trash answser items.
- *  1 when safely ignored and we must kept answer items
- */
 int act_resolution_error_cb(struct dns_requester *requester, int error_code)
 {
 	struct stream *stream;
