@@ -64,9 +64,11 @@ struct htx_sl *http_get_stline(struct htx *htx)
 {
 	struct htx_blk *blk;
 
+	BUG_ON(htx->first == -1);
 	blk = htx_get_first_blk(htx);
-	if (!blk || (htx_get_blk_type(blk) != HTX_BLK_REQ_SL && htx_get_blk_type(blk) != HTX_BLK_RES_SL))
+	if (!blk)
 		return NULL;
+	BUG_ON(htx_get_blk_type(blk) != HTX_BLK_REQ_SL && htx_get_blk_type(blk) != HTX_BLK_RES_SL);
 	return htx_get_blk_ptr(htx, blk);
 }
 
@@ -199,7 +201,7 @@ static int __http_find_header(const struct htx *htx, const void *pattern, struct
 					goto next_blk;
 				break;
 			case HTTP_FIND_FL_MATCH_SUB:
-				if (!strnistr(n.ptr, n.len, name.ptr, name.len))
+				if (strnistr(n.ptr, n.len, name.ptr, n.len) != NULL)
 					goto next_blk;
 				break;
 			default:
@@ -485,11 +487,11 @@ int http_replace_req_query(struct htx *htx, const struct ist query)
 /* Replace the response status in the HTX message <htx> by <status>. It returns
  * 1 on success, otherwise 0.
 */
-int http_replace_res_status(struct htx *htx, const struct ist status)
+int http_replace_res_status(struct htx *htx, const struct ist status, const struct ist reason)
 {
 	struct buffer *temp = get_trash_chunk();
 	struct htx_sl *sl = http_get_stline(htx);
-	struct ist vsn, reason;
+	struct ist vsn, r;
 
 	if (!sl)
 		return 0;
@@ -497,13 +499,15 @@ int http_replace_res_status(struct htx *htx, const struct ist status)
 	/* Start by copying old uri and version */
 	chunk_memcat(temp, HTX_SL_RES_VPTR(sl), HTX_SL_RES_VLEN(sl)); /* vsn */
 	vsn = ist2(temp->area, HTX_SL_RES_VLEN(sl));
-
-	chunk_memcat(temp, HTX_SL_RES_RPTR(sl), HTX_SL_RES_RLEN(sl)); /* reason */
-	reason = ist2(temp->area + vsn.len, HTX_SL_RES_RLEN(sl));
+	r = reason;
+	if (!isttest(r)) {
+		chunk_memcat(temp, HTX_SL_RES_RPTR(sl), HTX_SL_RES_RLEN(sl)); /* reason */
+		r = ist2(temp->area + vsn.len, HTX_SL_RES_RLEN(sl));
+	}
 
 	/* create the new start line */
 	sl->info.res.status = strl2ui(status.ptr, status.len);
-	return http_replace_stline(htx, vsn, status, reason);
+	return http_replace_stline(htx, vsn, status, r);
 }
 
 /* Replace the response reason in the HTX message <htx> by <reason>. It returns
@@ -965,17 +969,9 @@ int http_str_to_htx(struct buffer *buf, struct ist raw, char **errmsg)
 		goto error;
 	}
 	if ((flags & HTX_SL_F_CLEN) && h1m.body_len != (raw.len - ret)) {
-		struct ist clen = ist(ultoa(raw.len - ret));
-		int i;
-
-		memprintf(errmsg, "payload size does not match the announced content-length (%lu != %lu)."
-			  " C-L header is updated accordingly but it should be fixed to avoid any errors on future versions.",
-			  (unsigned long)(raw.len - ret), (unsigned long)h1m.body_len);
-
-		for (i = 0; hdrs[i].n.len; i++) {
-			if (isteqi(hdrs[i].n, ist("content-length")))
-				hdrs[i].v = clen;
-		}
+		memprintf(errmsg, "payload size does not match the announced content-length (%lu != %lu)",
+			  (raw.len - ret), h1m.body_len);
+		goto error;
 	}
 
 	htx = htx_from_buf(buf);
@@ -1396,7 +1392,7 @@ struct http_reply *http_parse_http_reply(const char **args, int *orig_arg, struc
 	struct stat stat;
 	const char *act_arg = NULL;
 	char *obj = NULL;
-	int cur_arg, cap = 0, objlen = 0, fd = -1;
+	int cur_arg, cap, objlen = 0, fd = -1;
 
 
 	reply = calloc(1, sizeof(*reply));
@@ -1410,12 +1406,10 @@ struct http_reply *http_parse_http_reply(const char **args, int *orig_arg, struc
 
 	if (px->conf.args.ctx == ARGC_HERR)
 		cap = (SMP_VAL_REQUEST | SMP_VAL_RESPONSE);
-	else {
-		if (px->cap & PR_CAP_FE)
-			cap |= ((px->conf.args.ctx == ARGC_HRQ) ? SMP_VAL_FE_HRQ_HDR : SMP_VAL_FE_HRS_HDR);
-		if (px->cap & PR_CAP_BE)
-			cap |= ((px->conf.args.ctx == ARGC_HRQ) ? SMP_VAL_BE_HRQ_HDR : SMP_VAL_BE_HRS_HDR);
-	}
+	else
+		cap = ((px->conf.args.ctx == ARGC_HRQ)
+		       ? ((px->cap & PR_CAP_FE) ? SMP_VAL_FE_HRQ_HDR : SMP_VAL_BE_HRQ_HDR)
+		       : ((px->cap & PR_CAP_BE) ? SMP_VAL_BE_HRS_HDR : SMP_VAL_FE_HRS_HDR));
 
 	cur_arg = *orig_arg;
 	while (*args[cur_arg]) {
