@@ -61,6 +61,79 @@ void ssl_sock_free_ssl_conf(struct ssl_bind_conf *conf)
 	}
 }
 
+/*
+ * Allocate and copy a ssl_bind_conf structure
+ */
+struct ssl_bind_conf *crtlist_dup_ssl_conf(struct ssl_bind_conf *src)
+{
+	struct ssl_bind_conf *dst;
+
+	if (!src)
+		return NULL;
+
+	dst = calloc(1, sizeof(*dst));
+	if (!dst)
+		return NULL;
+
+#if defined(OPENSSL_NPN_NEGOTIATED) && !defined(OPENSSL_NO_NEXTPROTONEG)
+	if (src->npn_str) {
+		dst->npn_str = strdup(src->npn_str);
+		if (!dst->npn_str)
+			goto error;
+	}
+#endif
+#ifdef TLSEXT_TYPE_application_layer_protocol_negotiation
+	if (src->alpn_str) {
+		dst->alpn_str = strdup(src->alpn_str);
+		if (!dst->alpn_str)
+			goto error;
+	}
+#endif
+	if (src->ca_file) {
+		dst->ca_file = strdup(src->ca_file);
+		if (!dst->ca_file)
+			goto error;
+	}
+	if (src->ca_verify_file) {
+		dst->ca_verify_file = strdup(src->ca_verify_file);
+		if (!dst->ca_verify_file)
+			goto error;
+	}
+	if (src->crl_file) {
+		dst->crl_file = strdup(src->crl_file);
+		if (!dst->crl_file)
+			goto error;
+	}
+	if (src->ciphers) {
+		dst->ciphers = strdup(src->ciphers);
+		if (!dst->ciphers)
+			goto error;
+	}
+#if (HA_OPENSSL_VERSION_NUMBER >= 0x10101000L)
+	if (src->ciphersuites) {
+		dst->ciphersuites = strdup(src->ciphersuites);
+		if (!dst->ciphersuites)
+			goto error;
+	}
+#endif
+	if (src->curves) {
+		dst->curves = strdup(src->curves);
+		if (!dst->curves)
+			goto error;
+	}
+	if (src->ecdhe) {
+		dst->ecdhe = strdup(src->ecdhe);
+		if (!dst->ecdhe)
+			goto error;
+	}
+	return dst;
+
+error:
+	ssl_sock_free_ssl_conf(dst);
+	free(dst);
+
+	return NULL;
+}
 
 /* free sni filters */
 void crtlist_free_filters(char **args)
@@ -123,6 +196,45 @@ void crtlist_entry_free(struct crtlist_entry *entry)
 	}
 	free(entry);
 }
+/*
+ * Duplicate a crt_list entry and its content (ssl_conf, filters/fcount)
+ * Return a pointer to the new entry
+ */
+struct crtlist_entry *crtlist_entry_dup(struct crtlist_entry *src)
+{
+	struct crtlist_entry *entry;
+
+	if (src == NULL)
+		return NULL;
+
+	entry = crtlist_entry_new();
+	if (entry == NULL)
+		return NULL;
+
+	if (src->filters) {
+		entry->filters = crtlist_dup_filters(src->filters, src->fcount);
+		if (!entry->filters)
+			goto error;
+	}
+	entry->fcount = src->fcount;
+	if (src->ssl_conf) {
+		entry->ssl_conf = crtlist_dup_ssl_conf(src->ssl_conf);
+		if (!entry->ssl_conf)
+			goto error;
+	}
+	entry->crtlist = src->crtlist;
+
+	return entry;
+
+error:
+
+	crtlist_free_filters(entry->filters);
+	ssl_sock_free_ssl_conf(entry->ssl_conf);
+	free(entry->ssl_conf);
+	free(entry);
+
+	return NULL;
+}
 
 /*
  * Allocate and initialize a crtlist_entry
@@ -199,7 +311,7 @@ struct crtlist *crtlist_new(const char *filename, int unique)
  *  <crt_path> is a ptr in <line>
  *  Return an error code
  */
-int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry, const char *file, int linenum, int from_cli, char **err)
+int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry, const char *file, int linenum, char **err)
 {
 	int cfgerr = 0;
 	int arg, newarg, cur_arg, i, ssl_b = 0, ssl_e = 0;
@@ -215,8 +327,8 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 		/* Check if we reached the limit and the last char is not \n.
 		 * Watch out for the last line without the terminating '\n'!
 		 */
-		memprintf(err, "line %d too long in file '%s', limit is %d characters",
-		          linenum, file, CRT_LINESIZE-1);
+		memprintf(err, "parsing [%s:%d]: line too long, limit is %d characters",
+		          file, linenum, CRT_LINESIZE-1);
 		cfgerr |= ERR_ALERT | ERR_FATAL;
 		goto error;
 	}
@@ -228,12 +340,12 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 			*line = 0;
 		} else if (*line == '[') {
 			if (ssl_b) {
-				memprintf(err, "too many '[' on line %d in file '%s'.", linenum, file);
+				memprintf(err, "parsing [%s:%d]: too many '['", file, linenum);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
 			if (!arg) {
-				memprintf(err, "file must start with a cert on line %d in file '%s'", linenum, file);
+				memprintf(err, "parsing [%s:%d]: file must start with a cert", file, linenum);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
@@ -242,12 +354,12 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 			*line = 0;
 		} else if (*line == ']') {
 			if (ssl_e) {
-				memprintf(err, "too many ']' on line %d in file '%s'.", linenum, file);
+				memprintf(err, "parsing [%s:%d]: too many ']'", file, linenum);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
 			if (!ssl_b) {
-				memprintf(err, "missing '[' in line %d in file '%s'.", linenum, file);
+				memprintf(err, "parsing [%s:%d]: missing '['", file, linenum);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
@@ -256,7 +368,7 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 			*line = 0;
 		} else if (newarg) {
 			if (arg == MAX_CRT_ARGS) {
-				memprintf(err, "too many args on line %d in file '%s'.", linenum, file);
+				memprintf(err, "parsing [%s:%d]: too many args ", file, linenum);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
@@ -283,17 +395,16 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 			goto error;
 		}
 	}
-
 	cur_arg = ssl_b ? ssl_b : 1;
 	while (cur_arg < ssl_e) {
 		newarg = 0;
 		for (i = 0; ssl_bind_kws[i].kw != NULL; i++) {
 			if (strcmp(ssl_bind_kws[i].kw, args[cur_arg]) == 0) {
 				newarg = 1;
-				cfgerr |= ssl_bind_kws[i].parse(args, cur_arg, NULL, ssl_conf, from_cli, err);
+				cfgerr |= ssl_bind_kws[i].parse(args, cur_arg, NULL, ssl_conf, err);
 				if (cur_arg + 1 + ssl_bind_kws[i].skip > ssl_e) {
-					memprintf(err, "ssl args out of '[]' for %s on line %d in file '%s'",
-					          args[cur_arg], linenum, file);
+					memprintf(err, "parsing [%s:%d]: ssl args out of '[]' for %s",
+					          file, linenum, args[cur_arg]);
 					cfgerr |= ERR_ALERT | ERR_FATAL;
 					goto error;
 				}
@@ -302,8 +413,8 @@ int crtlist_parse_line(char *line, char **crt_path, struct crtlist_entry *entry,
 			}
 		}
 		if (!cfgerr && !newarg) {
-			memprintf(err, "unknown ssl keyword %s on line %d in file '%s'.",
-				  args[cur_arg], linenum, file);
+			memprintf(err, "parsing [%s:%d]: unknown ssl keyword %s",
+				  file, linenum, args[cur_arg]);
 			cfgerr |= ERR_ALERT | ERR_FATAL;
 			goto error;
 		}
@@ -364,7 +475,7 @@ int crtlist_parse_file(char *file, struct bind_conf *bind_conf, struct proxy *cu
 		if (missing_lf != -1) {
 			memprintf(err, "parsing [%s:%d]: Stray NUL character at position %d.\n",
 			          file, linenum, (missing_lf + 1));
-			cfgerr |= ERR_WARN;
+			cfgerr |= ERR_ALERT | ERR_FATAL;
 			missing_lf = -1;
 			break;
 		}
@@ -375,8 +486,8 @@ int crtlist_parse_file(char *file, struct bind_conf *bind_conf, struct proxy *cu
 			/* Check if we reached the limit and the last char is not \n.
 			 * Watch out for the last line without the terminating '\n'!
 			 */
-			memprintf(err, "line %d too long in file '%s', limit is %d characters",
-				  linenum, file, (int)sizeof(thisline)-1);
+			memprintf(err, "parsing [%s:%d]: line too long, limit is %d characters",
+				  file, linenum, (int)sizeof(thisline)-1);
 			cfgerr |= ERR_ALERT | ERR_FATAL;
 			break;
 		}
@@ -400,7 +511,7 @@ int crtlist_parse_file(char *file, struct bind_conf *bind_conf, struct proxy *cu
 			goto error;
 		}
 
-		cfgerr |= crtlist_parse_line(thisline, &crt_path, entry, file, linenum, 0, err);
+		cfgerr |= crtlist_parse_line(thisline, &crt_path, entry, file, linenum, err);
 		if (cfgerr & ERR_CODE)
 			goto error;
 
@@ -413,8 +524,8 @@ int crtlist_parse_file(char *file, struct bind_conf *bind_conf, struct proxy *cu
 
 		if (*crt_path != '/' && global_ssl.crt_base) {
 			if ((strlen(global_ssl.crt_base) + 1 + strlen(crt_path)) > MAXPATHLEN) {
-				memprintf(err, "'%s' : path too long on line %d in file '%s'",
-					  crt_path, linenum, file);
+				memprintf(err, "parsing [%s:%d]: '%s' : path too long",
+					  file, linenum, crt_path);
 				cfgerr |= ERR_ALERT | ERR_FATAL;
 				goto error;
 			}
@@ -425,32 +536,82 @@ int crtlist_parse_file(char *file, struct bind_conf *bind_conf, struct proxy *cu
 		/* Look for a ckch_store or create one */
 		ckchs = ckchs_lookup(crt_path);
 		if (ckchs == NULL) {
-			if (stat(crt_path, &buf) == 0)
-				ckchs = ckchs_load_cert_file(crt_path, 0,  err);
-			else
-				ckchs = ckchs_load_cert_file(crt_path, 1,  err);
+			if (stat(crt_path, &buf) == 0) {
+
+				ckchs = ckchs_load_cert_file(crt_path, err);
+				if (ckchs == NULL) {
+					cfgerr |= ERR_ALERT | ERR_FATAL;
+					goto error;
+				}
+
+				entry->node.key = ckchs;
+				entry->crtlist = newlist;
+				ebpt_insert(&newlist->entries, &entry->node);
+				LIST_ADDQ(&newlist->ord_entries, &entry->by_crtlist);
+				LIST_ADDQ(&ckchs->crtlist_entry, &entry->by_ckch_store);
+
+			} else {
+				/* If we didn't find the file, this could be a
+				bundle, since 2.3 we don't support multiple
+				certificate in the same OpenSSL store, so we
+				emulate it by loading each file separately. To
+				do so we need to duplicate the entry in the
+				crt-list because it becomes independent */
+				char fp[MAXPATHLEN+1] = {0};
+				int n = 0;
+				struct crtlist_entry *entry_dup = entry; /* use the previous created entry */
+				for (n = 0; n < SSL_SOCK_NUM_KEYTYPES; n++) {
+					struct stat buf;
+					int ret;
+
+					ret = snprintf(fp, sizeof(fp), "%s.%s", path, SSL_SOCK_KEYTYPE_NAMES[n]);
+					if (ret > sizeof(fp))
+						continue;
+
+					ckchs = ckchs_lookup(fp);
+					if (!ckchs && stat(fp, &buf) == 0) {
+
+						ckchs = ckchs_load_cert_file(fp, err);
+						if (ckchs == NULL) {
+							cfgerr |= ERR_ALERT | ERR_FATAL;
+							goto error;
+						}
+
+						linenum++; /* we duplicate the line for this entry in the bundle */
+						if (!entry_dup) { /* if the entry was used, duplicate one */
+							linenum++;
+							entry_dup = crtlist_entry_dup(entry);
+							if (!entry_dup) {
+								cfgerr |= ERR_ALERT | ERR_FATAL;
+								goto error;
+							}
+							entry_dup->linenum = linenum;
+						}
+
+						entry_dup->node.key = ckchs;
+						entry_dup->crtlist = newlist;
+						ebpt_insert(&newlist->entries, &entry_dup->node);
+						LIST_ADDQ(&newlist->ord_entries, &entry_dup->by_crtlist);
+						LIST_ADDQ(&ckchs->crtlist_entry, &entry_dup->by_ckch_store);
+
+						entry_dup = NULL; /* the entry was used, we need a new one next round */
+					}
+				}
+			}
+		} else {
+			entry->node.key = ckchs;
+			entry->crtlist = newlist;
+			ebpt_insert(&newlist->entries, &entry->node);
+			LIST_ADDQ(&newlist->ord_entries, &entry->by_crtlist);
+			LIST_ADDQ(&ckchs->crtlist_entry, &entry->by_ckch_store);
 		}
-		if (ckchs == NULL) {
-			cfgerr |= ERR_ALERT | ERR_FATAL;
-			goto error;
-		}
-
-		if (cfgerr & ERR_CODE)
-			goto error;
-
-		entry->node.key = ckchs;
-		entry->crtlist = newlist;
-		ebpt_insert(&newlist->entries, &entry->node);
-		LIST_ADDQ(&newlist->ord_entries, &entry->by_crtlist);
-		LIST_ADDQ(&ckchs->crtlist_entry, &entry->by_ckch_store);
-
 		entry = NULL;
 	}
 
 	if (missing_lf != -1) {
 		memprintf(err, "parsing [%s:%d]: Missing LF on last line, file might have been truncated at position %d.\n",
 		          file, linenum, (missing_lf + 1));
-		cfgerr |= ERR_WARN;
+		cfgerr |= ERR_ALERT | ERR_FATAL;
 	}
 
 	if (cfgerr & ERR_CODE)
@@ -485,10 +646,6 @@ int crtlist_load_cert_dir(char *path, struct bind_conf *bind_conf, struct crtlis
 	char fp[MAXPATHLEN+1];
 	int cfgerr = 0;
 	struct ckch_store *ckchs;
-#if HA_OPENSSL_VERSION_NUMBER >= 0x1000200fL
-	int is_bundle;
-	int j;
-#endif
 
 	dir = crtlist_new(path, 1);
 	if (dir == NULL) {
@@ -528,59 +685,9 @@ int crtlist_load_cert_dir(char *path, struct bind_conf *bind_conf, struct crtlis
 				goto ignore_entry;
 			}
 
-#if HA_OPENSSL_VERSION_NUMBER >= 0x1000200fL
-			is_bundle = 0;
-			/* Check if current entry in directory is part of a multi-cert bundle */
-
-			if ((global_ssl.extra_files & SSL_GF_BUNDLE) && end) {
-				for (j = 0; j < SSL_SOCK_NUM_KEYTYPES; j++) {
-					if (!strcmp(end + 1, SSL_SOCK_KEYTYPE_NAMES[j])) {
-						is_bundle = 1;
-						break;
-					}
-				}
-
-				if (is_bundle) {
-					int dp_len;
-
-					dp_len = end - de->d_name;
-
-					/* increment i and free de until we get to a non-bundle cert
-					 * Note here that we look at de_list[i + 1] before freeing de
-					 * this is important since ignore_entry will free de. This also
-					 * guarantees that de->d_name continues to hold the same prefix.
-					 */
-					while (i + 1 < n && !strncmp(de_list[i + 1]->d_name, de->d_name, dp_len)) {
-						free(de);
-						i++;
-						de = de_list[i];
-					}
-
-					snprintf(fp, sizeof(fp), "%s/%.*s", path, dp_len, de->d_name);
-					ckchs = ckchs_lookup(fp);
-					if (ckchs == NULL)
-						ckchs = ckchs_load_cert_file(fp, 1,  err);
-					if (ckchs == NULL) {
-						free(de);
-						free(entry);
-						cfgerr |= ERR_ALERT | ERR_FATAL;
-						goto end;
-					}
-					entry->node.key = ckchs;
-					entry->crtlist = dir;
-					LIST_ADDQ(&ckchs->crtlist_entry, &entry->by_ckch_store);
-					LIST_ADDQ(&dir->ord_entries, &entry->by_crtlist);
-					ebpt_insert(&dir->entries, &entry->node);
-
-					/* Successfully processed the bundle */
-					goto ignore_entry;
-				}
-			}
-
-#endif
 			ckchs = ckchs_lookup(fp);
 			if (ckchs == NULL)
-				ckchs = ckchs_load_cert_file(fp, 0,  err);
+				ckchs = ckchs_load_cert_file(fp, err);
 			if (ckchs == NULL) {
 				free(de);
 				free(entry);
@@ -895,9 +1002,8 @@ static void cli_release_add_crtlist(struct appctx *appctx)
 {
 	struct crtlist_entry *entry = appctx->ctx.cli.p1;
 
-	if (entry) {
+	if (appctx->st2 != SETCERT_ST_FIN) {
 		struct ckch_inst *inst, *inst_s;
-
 		/* upon error free the ckch_inst and everything inside */
 		ebpt_delete(&entry->node);
 		LIST_DEL(&entry->by_crtlist);
@@ -911,9 +1017,8 @@ static void cli_release_add_crtlist(struct appctx *appctx)
 		free(entry->ssl_conf);
 		free(entry);
 	}
+
 	HA_SPIN_UNLOCK(CKCH_LOCK, &ckch_lock);
-	free(appctx->ctx.cli.err);
-	appctx->ctx.cli.err = NULL;
 }
 
 
@@ -930,23 +1035,30 @@ static int cli_io_handler_add_crtlist(struct appctx *appctx)
 	struct crtlist *crtlist = appctx->ctx.cli.p0;
 	struct crtlist_entry *entry = appctx->ctx.cli.p1;
 	struct ckch_store *store = entry->node.key;
+	struct buffer *trash = alloc_trash_chunk();
 	struct ckch_inst *new_inst;
+	char *err = NULL;
 	int i = 0;
 	int errcode = 0;
 
+	if (trash == NULL)
+		goto error;
+
 	/* for each bind_conf which use the crt-list, a new ckch_inst must be
 	 * created.
 	 */
 	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
-		goto end;
+		goto error;
 
 	while (1) {
 		switch (appctx->st2) {
 			case SETCERT_ST_INIT:
 				/* This state just print the update message */
-				chunk_printf(&trash, "Inserting certificate '%s' in crt-list '%s'", store->path, crtlist->node.key);
-				if (ci_putchk(si_ic(si), &trash) == -1)
+				chunk_printf(trash, "Inserting certificate '%s' in crt-list '%s'", store->path, crtlist->node.key);
+				if (ci_putchk(si_ic(si), trash) == -1) {
+					si_rx_room_blk(si);
 					goto yield;
+				}
 				appctx->st2 = SETCERT_ST_GEN;
 				/* fallthrough */
 			case SETCERT_ST_GEN:
@@ -957,39 +1069,28 @@ static int cli_io_handler_add_crtlist(struct appctx *appctx)
 					struct bind_conf *bind_conf = bind_conf_node->bind_conf;
 					struct sni_ctx *sni;
 
-					appctx->ctx.cli.p2 = bind_conf_node;
-
 					/* yield every 10 generations */
 					if (i > 10) {
-						si_rx_endp_more(si); /* let's come back later */
+						appctx->ctx.cli.p2 = bind_conf_node;
 						goto yield;
 					}
 
-					/* display one dot for each new instance */
-					if (ci_putstr(si_ic(si), ".") == -1)
-						goto yield;
-
 					/* we don't support multi-cert bundles, only simple ones */
-					appctx->ctx.cli.err = NULL;
-					errcode |= ckch_inst_new_load_store(store->path, store, bind_conf, entry->ssl_conf, entry->filters, entry->fcount, &new_inst, &appctx->ctx.cli.err);
-					if (errcode & ERR_CODE) {
-						appctx->st2 = SETCERT_ST_ERROR;
+					errcode |= ckch_inst_new_load_store(store->path, store, bind_conf, entry->ssl_conf, entry->filters, entry->fcount, &new_inst, &err);
+					if (errcode & ERR_CODE)
 						goto error;
-					}
 
 					/* we need to initialize the SSL_CTX generated */
 					/* this iterate on the newly generated SNIs in the new instance to prepare their SSL_CTX */
 					list_for_each_entry(sni, &new_inst->sni_ctx, by_ckch_inst) {
 						if (!sni->order) { /* we initialized only the first SSL_CTX because it's the same in the other sni_ctx's */
-							appctx->ctx.cli.err = NULL;
-							errcode |= ssl_sock_prepare_ctx(bind_conf, new_inst->ssl_conf, sni->ctx, &appctx->ctx.cli.err);
-							if (errcode & ERR_CODE) {
-								appctx->st2 = SETCERT_ST_ERROR;
+							errcode |= ssl_sock_prepare_ctx(bind_conf, new_inst->ssl_conf, sni->ctx, &err);
+							if (errcode & ERR_CODE)
 								goto error;
-							}
 						}
 					}
-
+					/* display one dot for each new instance */
+					chunk_appendf(trash, ".");
 					i++;
 					LIST_ADDQ(&store->ckch_inst, &new_inst->by_ckchs);
 					LIST_ADDQ(&entry->ckch_inst, &new_inst->by_crtlist_entry);
@@ -998,52 +1099,46 @@ static int cli_io_handler_add_crtlist(struct appctx *appctx)
 				appctx->st2 = SETCERT_ST_INSERT;
 				/* fallthrough */
 			case SETCERT_ST_INSERT:
-				/* the insertion is called for every instance of the store, not
-				 * only the one we generated.
-				 * But the ssl_sock_load_cert_sni() skip the sni already
-				 * inserted. Not every instance has a bind_conf, it could be
-				 * the store of a server so we should be careful */
-
+				/* insert SNIs in bind_conf */
 				list_for_each_entry(new_inst, &store->ckch_inst, by_ckchs) {
-					if (!new_inst->bind_conf) /* this is a server instance */
-						continue;
 					HA_RWLOCK_WRLOCK(SNI_LOCK, &new_inst->bind_conf->sni_lock);
 					ssl_sock_load_cert_sni(new_inst, new_inst->bind_conf);
 					HA_RWLOCK_WRUNLOCK(SNI_LOCK, &new_inst->bind_conf->sni_lock);
 				}
 				entry->linenum = ++crtlist->linecount;
-				appctx->ctx.cli.p1 = NULL;
-				appctx->st2 = SETCERT_ST_SUCCESS;
-				/* fallthrough */
-			case SETCERT_ST_SUCCESS:
-				chunk_reset(&trash);
-				chunk_appendf(&trash, "\n");
-				if (appctx->ctx.cli.err)
-					chunk_appendf(&trash, "%s", appctx->ctx.cli.err);
-				chunk_appendf(&trash, "Success!\n");
-				if (ci_putchk(si_ic(si), &trash) == -1)
-					goto yield;
 				appctx->st2 = SETCERT_ST_FIN;
 				goto end;
-
-			case SETCERT_ST_ERROR:
-			  error:
-				chunk_printf(&trash, "\n%sFailed!\n", appctx->ctx.cli.err);
-				if (ci_putchk(si_ic(si), &trash) == -1)
-					goto yield;
-				goto end;
-
-			default:
-				goto end;
 		}
 	}
 
 end:
+	chunk_appendf(trash, "\n");
+	if (errcode & ERR_WARN)
+		chunk_appendf(trash, "%s", err);
+	chunk_appendf(trash, "Success!\n");
+	if (ci_putchk(si_ic(si), trash) == -1)
+		si_rx_room_blk(si);
+	free_trash_chunk(trash);
 	/* success: call the release function and don't come back */
 	return 1;
 yield:
-	si_rx_room_blk(si);
+	/* store the state */
+	if (ci_putchk(si_ic(si), trash) == -1)
+		si_rx_room_blk(si);
+	free_trash_chunk(trash);
+	si_rx_endp_more(si); /* let's come back later */
 	return 0; /* should come back */
+
+error:
+	/* spin unlock and free are done in the release function */
+	if (trash) {
+		chunk_appendf(trash, "\n%sFailed!\n", err);
+		if (ci_putchk(si_ic(si), trash) == -1)
+			si_rx_room_blk(si);
+		free_trash_chunk(trash);
+	}
+	/* error: call the release function and don't come back */
+	return 1;
 }
 
 
@@ -1102,7 +1197,7 @@ static int cli_parse_add_crtlist(char **args, char *payload, struct appctx *appc
 			goto error;
 		}
 		/* cert_path is filled here */
-		cfgerr |= crtlist_parse_line(payload, &cert_path, entry, "CLI", 1, 1, &err);
+		cfgerr |= crtlist_parse_line(payload, &cert_path, entry, "CLI", 1, &err);
 		if (cfgerr & ERR_CODE)
 			goto error;
 	} else {
@@ -1148,10 +1243,6 @@ static int cli_parse_add_crtlist(char **args, char *payload, struct appctx *appc
 		memprintf(&err, "certificate '%s' does not exist!", cert_path);
 		goto error;
 	}
-	if (store->multi) {
-		memprintf(&err, "certificate '%s' is a bundle. You can disable the bundle merging with the directive 'ssl-load-extra-files' in the global section.", cert_path);
-		goto error;
-	}
 	if (store->ckch == NULL || store->ckch->cert == NULL) {
 		memprintf(&err, "certificate '%s' is empty!", cert_path);
 		goto error;
@@ -1203,7 +1294,6 @@ static int cli_parse_del_crtlist(char **args, char *payload, struct appctx *appc
 	int linenum = 0;
 	char *colons;
 	char *end;
-	int error_message_dumped = 0;
 
 	if (!cli_has_level(appctx, ACCESS_LVL_ADMIN))
 		return 1;
@@ -1247,10 +1337,6 @@ static int cli_parse_del_crtlist(char **args, char *payload, struct appctx *appc
 		memprintf(&err, "certificate '%s' does not exist!", cert_path);
 		goto error;
 	}
-	if (store->multi) {
-		memprintf(&err, "certificate '%s' is a bundle. You can disable the bundle merging with the directive 'ssl-load-extra-files' in the global section.", cert_path);
-		goto error;
-	}
 	if (store->ckch == NULL || store->ckch->cert == NULL) {
 		memprintf(&err, "certificate '%s' is empty!", cert_path);
 		goto error;
@@ -1288,20 +1374,6 @@ static int cli_parse_del_crtlist(char **args, char *payload, struct appctx *appc
 		goto error;
 	}
 
-	/* Iterate over all the instances in order to see if any of them is a
-	 * default instance. If this is the case, the entry won't be suppressed. */
-	list_for_each_entry_safe(inst, inst_s, &entry->ckch_inst, by_crtlist_entry) {
-		if (inst->is_default && !inst->bind_conf->strict_sni) {
-			if (!error_message_dumped) {
-				memprintf(&err, "certificate '%s' cannot be deleted, it is used as default certificate by the following frontends:\n", cert_path);
-				error_message_dumped = 1;
-			}
-			memprintf(&err, "%s\t- %s:%d\n", err, inst->bind_conf->file, inst->bind_conf->line);
-		}
-	}
-	if (error_message_dumped)
-		goto error;
-
 	/* upon error free the ckch_inst and everything inside */
 
 	ebpt_delete(&entry->node);
